package cz.fidentis.analyst.visitors.kdtree;

import cz.fidentis.analyst.kdtree.KdNode;
import cz.fidentis.analyst.kdtree.KdTreeVisitor;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import java.util.LinkedList;
import java.util.Queue;
import javax.vecmath.Vector3d;

/**
 * Abstract super-class for visitors that find nodes of k-d trees that are closest to the given 3D location.
 * <p>
 * A single visitor can be applied to multiple k-d trees.
 * <ul>
 * <li>If one tree has a closer distance to the given 3D point than all other trees,
 * then only this tree is taken into consideration. If there are multiple trees
 * with the same minimal distance, then only one is taken into account.</li>
 * <li>From the one closets tree, only one closest node is returned. 
 * If there are multiple nodes with the same minimal distance then only (random)
 * one is selected and returned.</li>
 * </ul>
 * </p>
 * <p>
 * This visitor is thread-safe. 
 * </p>
 * 
 * @author Daniel Schramm
 */
public abstract class ClosestNodeAbstractVisitor extends KdTreeVisitor {

    private final Vector3d myPoint;
    private double distance = Double.POSITIVE_INFINITY;

    /**
     * @param point The 3D location for which the nodes are to be computed.
     * @param asynchronous If {@code true}, then asynchronous visitor is created.
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public ClosestNodeAbstractVisitor(Vector3d point, boolean asynchronous) {
        super(asynchronous);
        if (point == null) {
            throw new IllegalArgumentException("point");
        }
        this.myPoint = point;
    }

    /**
     * Returns the minimal distance
     * @return the minimal distance
     */
    public double getDistance() {
        return distance;
    }
    
    /**
     * Returns the 3D point from which the distance is computed
     * @return the 3D point from which the distance is computed
     */
    protected Vector3d getPoint3d() {
        return myPoint;
    }
    
    /**
     * Finds the closest node for given 3D location.
     *
     * @return the closest node of the kd-tree or null
     */
    protected KdNode closestNode(KdNode root) {
        if (root == null) {
            return null;
        }
        
        /*
         * First, find the closest node
         */
        double minDistance = Double.POSITIVE_INFINITY;
        KdNode nearNode = root;
        KdNode searchedNode = root;

        while (searchedNode != null) {
            Vector3d nodePos = searchedNode.getLocation();
            double dist = MeshPoint.distance(myPoint, nodePos);
            if (dist < minDistance) {
                nearNode = searchedNode;
                minDistance = dist;
            }

            if (firstIsLessThanSecond(myPoint, nodePos, searchedNode.getDepth())) {
                searchedNode = searchedNode.getLesser();
            } else {
                searchedNode = searchedNode.getGreater();
            }
        }

        /*
         * Second, search for a vertex that could be potentially closer than
         * the nearest vertex already found
         */
        double distOnAxis;
        Queue<KdNode> queue = new LinkedList<>();
        queue.add(root);

        while (!queue.isEmpty()) {
            
            if (minDistance == 0) { // nothing can be closer
                break;
            }

            searchedNode = queue.poll();
            Vector3d nodePos = searchedNode.getLocation();

            double dist = MeshPoint.distance(myPoint, nodePos);
            if (dist < minDistance) {
                nearNode = searchedNode;
                minDistance = dist;
            }

            distOnAxis = minDistanceIntersection(nodePos, myPoint, searchedNode.getDepth());

            if (distOnAxis > minDistance) {
                if (firstIsLessThanSecond(myPoint, nodePos, searchedNode.getDepth())) {
                    if (searchedNode.getLesser() != null) {
                        queue.add(searchedNode.getLesser());
                    }
                } else {
                    if (searchedNode.getGreater() != null) {
                        queue.add(searchedNode.getGreater());
                    }
                }
            } else {
                if (searchedNode.getLesser() != null) {
                    queue.add(searchedNode.getLesser());
                }
                if (searchedNode.getGreater() != null) {
                    queue.add(searchedNode.getGreater());
                }
            }
        }
        
        return nearNode;
    }
    
    protected boolean firstIsLessThanSecond(Vector3d v1, Vector3d v2, int level){
        switch (level % 3) {
            case 0:
                return v1.x <= v2.x;
            case 1:
                return v1.y <= v2.y;
            case 2:
                return v1.z <= v2.z;
            default:
                break;
        }
        return false;
    }
    
    /**
     * Calculates distance between two points
     * (currently searched node and point to which we want to find nearest neighbor)
     * (based on axis)
     *
     */
    protected double minDistanceIntersection(Vector3d nodePosition, Vector3d pointPosition, int level){
        switch (level % 3) {
            case 0:
                return Math.abs(nodePosition.x - pointPosition.x);
            case 1:
                return Math.abs(nodePosition.y - pointPosition.y);
            default:
                return Math.abs(nodePosition.z - pointPosition.z);
        }
    }
    
    protected boolean updateMinimalDistance(double dist) {
        if (dist< distance) {
            distance = dist;
            return true;
        } else {
            return false;
        }
    }
}
