package cz.fidentis.analyst.visitors.kdtree;

import cz.fidentis.analyst.kdtree.KdNode;
import cz.fidentis.analyst.kdtree.KdTree;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import javax.vecmath.Vector3d;

/**
 * This visitor finds <b>a single k-d tree node</b>,  which is closest to the given 3D location.
 * <p>
 * A single visitor can be applied to multiple k-d trees.
 * <ul>
 * <li>If one tree has a closer distance to the given 3D point than all other trees,
 * then only this tree is taken into consideration. If there are multiple trees
 * with the same minimal distance, then only one is taken into account.</li>
 * <li>From the closest tree, <b>only one closest node is returned</b>. 
 * If there are multiple nodes with the same minimal distance then only (random)
 * one is selected and returned.</li>
 * </ul>
 * </p>
 * <p>
 * This visitor is thread-safe. 
 * </p>
 * 
 * @author Daniel Schramm
 */
public class ClosestNodeVisitor extends ClosestNodeAbstractVisitor {
    
    private KdNode closestNode;
    
    /**
     * @param point The 3D location for which the closest points are to be computed.
     * @param asynchronous If {@code true}, then asynchronous visitor is created.
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public ClosestNodeVisitor(Vector3d point, boolean asynchronous) {
        super(point, asynchronous);
    }

    /**
     * Returns the closest node of the visited KD tree.
     * 
     * @return closest node of the visited KD tree
     */
    public KdNode getClosestNode() {
        return closestNode;
    }
    
    @Override
    protected void visitKdTree(KdTree kdTree) {
        KdNode foundNode = closestNode(kdTree.getRoot());
        final double foundDistance = MeshPoint.distance(getPoint3d(), foundNode.getLocation());
        synchronized (this) {
            if (updateMinimalDistance(foundDistance)) {
                closestNode = foundNode;
            }
        }
    }
}
