package cz.fidentis.analyst.visitors.kdtree;

import cz.fidentis.analyst.kdtree.KdNode;
import cz.fidentis.analyst.kdtree.KdTree;
import cz.fidentis.analyst.kdtree.KdTreeVisitor;
import cz.fidentis.analyst.visitors.Distance;
import javax.vecmath.Vector3d;

/**
 * This visitor finds the minimal distance between the given 3D point and points 
 * (e.g., mesh vertices) stored in k-d trees.
 * <p>
 * This visitor is thread-safe. 
 * </p>
 * 
 * @author Daniel Schramm
 * @author Radek Oslejsek
 */
public class KdTreeDistance extends KdTreeVisitor implements Distance {
    
    private double distance = Double.POSITIVE_INFINITY;
    private Vector3d point3d;
    
    /**
     * Constructor.
     * @param point A 3D point from which distance is computed. Must not be {@code null}
     * @param asynchronous If {@code true}, then asynchronous visitor is created.
     * @throws {@code IllegalArgumentException} if some parametr is wrong
     */
    public KdTreeDistance(Vector3d point, boolean asynchronous) {
        super(asynchronous);
        if (point == null) {
            throw new IllegalArgumentException("point");
        }
        this.point3d = point;
    }

    @Override
    public double getDistance() {
        return distance;
    }

    @Override
    protected void visitKdTree(KdTree kdTree) {
        final ClosestNodeVisitor visitor = new ClosestNodeVisitor(point3d, false); 
        kdTree.accept(visitor);
        final KdNode closestNode = visitor.getClosestNode();
        final Vector3d aux = new Vector3d(closestNode.getLocation());
        aux.sub(point3d);
        final double dist = aux.length();
        
        synchronized (this) {
            if (dist < distance) {
                distance = dist;
            }
        }
    }

}
