package cz.fidentis.analyst.visitors.kdtree;

import cz.fidentis.analyst.kdtree.KdNode;
import cz.fidentis.analyst.kdtree.KdTree;
import cz.fidentis.analyst.kdtree.KdTreeVisitor;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.visitors.DistanceWithNearestPoints;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import javax.vecmath.Vector3d;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * This visitor finds the minimal distance between the given 3D point and and points 
 * (e.g., mesh vertices) stored in k-d trees. 
 * All closest mesh facets and all their closests vertices are returned.
 * <p>
 * This visitor is thread-safe. 
 * </p>
 * 
 * @author Daniel Schramm
 * @author Radek Oslejsek
 */
public class KdTreeDistanceToVertices extends KdTreeVisitor implements DistanceWithNearestPoints {
    
    private double distance = Double.POSITIVE_INFINITY;
    private final Vector3d point3d;
    private Map<MeshFacet, List<Vector3d>> nearestPoints = new HashMap<>();
    
    /**
     * @param point A 3D point from which distance is computed. Must not be {@code null}
     * @param asynchronous If {@code true}, then asynchronous visitor is created.
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public KdTreeDistanceToVertices(Vector3d point, boolean asynchronous) {
        super(asynchronous);
        if (point == null) {
            throw new IllegalArgumentException("point");
        }
        this.point3d = point;
    }
    
    @Override
    public Map<MeshFacet, List<Vector3d>> getNearestPoints() {
        return Collections.unmodifiableMap(nearestPoints);
    }

    @Override
    public double getDistance() {
        return distance;
    }
    
    @Override
    protected void visitKdTree(KdTree kdTree) {
        final ClosestNodeMultipleVisitor visitor = new ClosestNodeMultipleVisitor(point3d, false); 
        kdTree.accept(visitor);
        final Set<KdNode> closestNodes = visitor.getClosestNodes();
        for (final KdNode node: closestNodes) { // For all nodes
            final Vector3d aux = new Vector3d(node.getLocation());
            aux.sub(point3d);
            final double dist = aux.length();
        
            synchronized (this) {
                if (dist > distance) {
                    continue;
                }
                if (dist < distance) {
                    distance = dist;
                    nearestPoints.clear();                    
                }
                for (Entry<MeshFacet, Integer> entry: node.getFacets().entrySet()) {
                    MeshFacet facet = entry.getKey();
                    MeshPoint point = facet.getVertex(entry.getValue());
                    nearestPoints.putIfAbsent(facet, new ArrayList<>());
                    nearestPoints.get(facet).add(point.getPosition());
                }
            }
        }
    }
}
