package cz.fidentis.analyst.visitors.kdtree;

import cz.fidentis.analyst.kdtree.KdNode;
import cz.fidentis.analyst.kdtree.KdTree;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshFacetImpl;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.mesh.core.MeshPointImpl;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import javax.vecmath.Vector3d;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.assertTrue;


/**
 * @author Maria Kocurekova
 */
public class ClosestNodeVisitorTest {

    private Vector3d position = new Vector3d(0.1f, 0.5f, 0.7f);
    private Vector3d normalAndTextCoord = new Vector3d(0,0,0);
    
    protected boolean containsPoint(KdNode node, MeshPoint p) {
        if (node == null) {
            return false;
        }
        if (node.getLocation().equals(p.getPosition())) {
            return true;
        }
        if (containsPoint(node.getLesser(), p)) {
            return true;
        }
        if (containsPoint(node.getGreater(), p)) {
            return true;
        }
        return false;
    }
    
    /**
     * Returns the closest point
     *
     * @param points list of points
     * @param p point to which the closest point is searched
     * @return the closest point to p
     */
    private MeshPoint findClosestSequential(List<MeshPoint> points, MeshPoint p){
        if(points.isEmpty()){
            return null;
        }

        MeshPoint closest = points.get(0);
        double dis = closest.distanceTo(p);

        for(MeshPoint cl : points){
            if(cl.distanceTo(p) < dis){
                dis = cl.distanceTo(p);
                closest = cl;
            }
        }

        return closest;
    }
    


    @Test
    public void testCountOfFacets(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i = 0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        
        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        assertTrue(containsPoint(tree.getRoot(), p));
        
        ClosestNodeVisitor vis = new ClosestNodeVisitor(p.getPosition(), false);
        tree.accept(vis);
        KdNode node = vis.getClosestNode();
        assertNotNull(node);
        assertEquals(2, node.getFacets().size());
    }

    @Test
    public void testFindClosestNotInTreeTwoFacets(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 5; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i =5; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }


        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);

        MeshPoint p = new MeshPointImpl(new Vector3d(2.5f, 0.3f, 1.2f), normalAndTextCoord,normalAndTextCoord);
        ClosestNodeVisitor vis = new ClosestNodeVisitor(p.getPosition(), false);
        tree.accept(vis);
        KdNode node = vis.getClosestNode();
        assertNotNull(node);
        
        MeshPoint seq = findClosestSequential(points, p);
        assertNotNull(seq);
        
        assertEquals(node.getLocation(), seq.getPosition());
    }
    @Test
    public void testIndexFoundTwoFacetsNoDuplicates(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 5; i < 20; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i =0; i < 5; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }


        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        
        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        ClosestNodeVisitor vis = new ClosestNodeVisitor(p.getPosition(), false);
        tree.accept(vis);
        KdNode node = vis.getClosestNode();
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);        
    }

    @Test
    public void testIndexFoundTwoFacetsAllDuplicates(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i =0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }


        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        
        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        ClosestNodeVisitor vis = new ClosestNodeVisitor(p.getPosition(), false);
        tree.accept(vis);
        KdNode node = vis.getClosestNode();
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);        

        MeshFacet second = (new ArrayList<>(found.keySet())).get(1);
        int secondIndex = (new ArrayList<>(found.values())).get(1);
        assertEquals(second.getVertex(secondIndex), p);
        
        assertEquals(0.0f, second.getVertex(secondIndex).distanceTo(entry.getKey().getVertex(entry.getValue())));
    }


    @Test
    public void testOneFacetNearNeighbourIndex(){
        MeshFacet facet1 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 1; i < 20; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }

        KdTree tree = new KdTree(facet1);
        
        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        ClosestNodeVisitor vis = new ClosestNodeVisitor(p.getPosition(), false);
        tree.accept(vis);
        KdNode node = vis.getClosestNode();
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);
    }

    @Test
    public void testSetOfPointsNearNeighbourIndex(){
        Set<MeshPoint> points = new HashSet<>();
        List<MeshPoint> pointsList = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 1; i < 20; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            pointsList.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }

        KdTree tree = new KdTree(points);
        
        Random r = new Random();
        MeshPoint p = pointsList.get(r.nextInt(points.size()));
        ClosestNodeVisitor vis = new ClosestNodeVisitor(p.getPosition(), false);
        tree.accept(vis);
        KdNode node = vis.getClosestNode();
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);
    }
    
}
